/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
import UIKit
import BoxContentSDK

class BoxContentViewController: UIViewController, UITextFieldDelegate
{
    // MARK: Properties
    @IBOutlet weak var logView: UITextView!
    @IBOutlet weak var downloadButton: UIButton!
    @IBOutlet weak var fileField: UITextField!
    @IBOutlet weak var logLabel: UILabel!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    @IBOutlet weak var dirLabel: UILabel!
    
    // MARK: Variables
    var referencedController: FirmwareSourceViewController!
    var boxClient: BOXContentClient!
    var outStream: OutputStream!
    
    // MARK: Actions
    @IBAction func downloadPressed(_ sender: UIButton)
    {
        if (fileField.text?.isEmpty)!
        {
            logView.text = logView.text + "\n Filename cannot be empty!"
            enableGUIElements(state: true)
            return
        }
        
        /* Searching for the file */
        enableGUIElements(state: false)
        
        let boxClient = BOXContentClient.default()
        let searchRequest = boxClient?.searchRequest(withQuery: fileField.text, in: NSMakeRange(0, 10))
        
        searchRequest?.perform(completion:
            {
                (items: [BOXItem]?, count: UInt, range: NSRange, error: Error?) in
                
                if(error != nil)
                {
                    self.logView.text = self.logView.text +  "\nError searching for file!"
                    self.enableGUIElements(state: true)
                    return
                }
                else if (items?.count == 0 || items?[0].name != self.fileField.text)
                {
                    self.logView.text = self.logView.text +  "\nNo files found named " + self.fileField.text!
                    self.enableGUIElements(state: true)
                    return
                }
                
                self.logView.text = self.logView.text + "\nFile " + (items?[0].name)! + " found!"
                
                let downloadReq = boxClient?.fileDownloadRequest(withID: items![0].modelID, to: self.outStream)
                downloadReq?.perform(progress: nil, completion:
                    { (error: Error?) in
                    
                        if(error != nil)
                        {
                            self.logView.text = self.logView.text +  "\nError downloading file!"
                            self.enableGUIElements(state: true)
                            return
                        }
                        
                        let data = self.outStream.property(forKey: .dataWrittenToMemoryStreamKey) as! Data
                        
                        let importedImage: ImportedFirmwareImage = ImportedFirmwareImage(firmware: data,
                                                                                         source: DownloadSource.FIRMARE_BOXDRIVE,
                                                                                         filename: self.fileField.text!)
                        
                        self.referencedController.referencedController.curFirmware = importedImage
                        self.referencedController.referencedController.firmwareNameLabel.text = importedImage.getTableLabel()
                        self.referencedController.referencedController.firmwareImages.append(importedImage)
                        self.referencedController.referencedController.saveFirmwareImages()
                        
                        self.navigationController!.popToViewController(self.referencedController.referencedController,
                                                                       animated: true)
                })
                
                
                
            })
    }
    
    // MARK: Functions
    override func viewDidLoad()
    {
        super.viewDidLoad()
        self.boxClient = BOXContentClient.default()
        fileField.delegate = self
        self.outStream = OutputStream(toMemory: ())
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(BoxContentViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        dirLabel.adjustsFontSizeToFitWidth = true
        
        downloadButton.layer.cornerRadius = 8.0
        downloadButton.clipsToBounds = true
        
        viewRotated()
    }
    
    func enableGUIElements(state: Bool)
    {
        fileField.isEnabled = state
        downloadButton.isEnabled = state
        
    }
    
    func showAlert(title : String, message: String)
    {
        let alert = UIAlertController(
            title: title,
            message: message,
            preferredStyle: UIAlertControllerStyle.alert
        )
        let ok = UIAlertAction(
            title: "OK",
            style: UIAlertActionStyle.default,
            handler: nil
        )
        alert.addAction(ok)
        present(alert, animated: true, completion: nil)
    }
    
    override func viewDidAppear(_ animated: Bool)
    {
        BOXContentClient.default().authenticate(completionBlock:
            {
                (user: BOXUser?, error: Error?) -> Void in
                
                if(error == nil)
                {
                    self.logView.text = "Box Drive successfully authenticated!"
                    self.enableGUIElements(state: true)
                }
                else
                {
                    self.navigationController!.popToViewController(self.referencedController.referencedController,
                                                                   animated: true)
                }
                
        })
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.logLabel.isHidden = false
                    self.logView.isHidden = false
                }
                else if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.logLabel.isHidden = true
                    self.logView.isHidden = true
                }
        }
    }
    
    @objc func keyboardNotification(notification: NSNotification)
    {
        
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }

    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        fileField.resignFirstResponder()
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {
        
    }
}
