/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import CoreBluetooth

class DataServiceViewController: UIViewController, UIPickerViewDelegate, UITextFieldDelegate
{

    // MARK: - Properties
    @IBOutlet weak var dataLog: UITextView!
    @IBOutlet weak var writeButton: UIButton!
    @IBOutlet weak var charPicker: UIPickerView!
    @IBOutlet weak var logLabel: UILabel!
    @IBOutlet weak var charLabel: UILabel!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    @IBOutlet weak var inputField: UITextField!

    // MARK: Variables
    let hexSet = NSCharacterSet(charactersIn: "0123456789ABCDEFabcdef").inverted
    
    // MARK: - Actions
    @IBAction func writeButtonPressed(_ sender: Any)
    {
        if(inputField.text?.isEmpty)!
        {
            return
        }
        
        if(inputField.isFirstResponder)
        {
            inputField.resignFirstResponder()
        }
        
        if(charPicker.selectedRow(inComponent: 0) == 0)
        {
            let dataArr: [UInt8] = Array(inputField.text!.utf8)
            let writeArray = Data(dataArr)
            referencedController.btDevice.writeValue(writeArray, for: referencedController.dataCharString,
                                                     type: CBCharacteristicWriteType.withResponse)
            printLogMessage(message: "Value of \(inputField.text!) written to WRITE characteristic")
        }
        else
        {
            let dataArray: [UInt8] = [UInt8](toByteArray(UInt(inputField.text!, radix: 16)))
            let writeArray = Data(dataArray)
            referencedController.btDevice.writeValue(writeArray, for: referencedController.dataCharData,
                                                     type: CBCharacteristicWriteType.withoutResponse)
            
            printLogMessage(message: "Value of \(inputField.text!) written to WRITE_NO_RESP characteristic")
        }
    }
    
    func printLogMessage(message: String)
    {
        self.dataLog.text = self.dataLog.text + self.dateFormatter.string(from: Date()) + ":    " + message + "\n"
        let range = NSRange(location: self.dataLog.text.characters.count-1,length: 0)
        self.dataLog.scrollRangeToVisible(range)
    }
    
    // MARK: - Variables
    var referencedController: ProjectZeroTabBarController!
    let dateFormatter = DateFormatter()
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        referencedController = tabBarController as! ProjectZeroTabBarController
        referencedController.dataServiceController = self
        charPicker.delegate = self
        inputField.delegate = self
        
        /* Making the GUI look curvy */
        logLabel.clipsToBounds = true
        logLabel.layer.cornerRadius = 8.0
        writeButton.clipsToBounds = true
        writeButton.layer.cornerRadius = 8.0
        charLabel.clipsToBounds = true
        charLabel.layer.cornerRadius = 8.0
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(DataServiceViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
        dateFormatter.dateFormat = "HH:mm:ss"
        
        dataLog.text = dataLog.text + "\n"
        
        viewRotated()

    }
    
    @objc func keyboardNotification(notification: NSNotification)
    {
        
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
    }
    
    // MARK: UIPickerViewDelegate
    func numberOfComponentsInPickerView(_ pickerView: UIPickerView) -> Int
    {
        return 1
    }
    
    // The number of rows of data
    func pickerView(_ pickerView: UIPickerView, numberOfRowsInComponent component: Int) -> Int
    {
        return 2
    }
    
    func pickerView(_ pickerView: UIPickerView, viewForRow row: Int,
                    forComponent component: Int, reusing view: UIView?) -> UIView
    {
        let pickerLabel = UILabel()
        
        if(row == 0)
        {
            pickerLabel.text = "Write String"
        }
        else
        {
            pickerLabel.text = "Write without Response"
        }
        
        return pickerLabel;
    }
    
    func pickerView(_ pickerView: UIPickerView, didSelectRow row: Int,
                    inComponent component: Int)
    {
        inputField.text = ""
    }
    
    // MARK: UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        inputField.resignFirstResponder()
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {

    }
    
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange,
                   replacementString string: String) -> Bool
    {
        if(charPicker.selectedRow(inComponent: 0) == 0)
        {
            if((string.characters.count + (inputField.text?.characters.count)!) <= 20)
            {
                return true
            }
            else
            {
                return false
            }
        }
        else
        {
            if((string.characters.count + (inputField.text?.characters.count)!) > 8)
            {
                return false
            }
            
            let separatedChar = string.components(separatedBy: hexSet)
            let filteredNums = separatedChar.joined(separator: "")
            
            if (filteredNums != string)
            {
                return false
            }
            
            return true
        }
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.dataLog.isHidden = false
                    self.logLabel.isHidden = false
                }
                else if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.dataLog.isHidden = true
                    self.logLabel.isHidden = true
                }
        }
    }
    
}
