/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import CoreBluetooth

class OADViewController: UIViewController, CBPeripheralDelegate, UITextFieldDelegate
{

    // MARK: - Properties
    @IBOutlet weak var sourceLabel: UIView!
    @IBOutlet weak var firmwareSourceView: UIView!
    @IBOutlet weak var firmwareNameLabel: UILabel!
    @IBOutlet weak var firmwareOptionsView: UIView!
    @IBOutlet weak var firmwareOptionsViewLabel: UIView!
    @IBOutlet weak var sapLabel: UILabel!
    @IBOutlet weak var snpLabel: UILabel!
    @IBOutlet weak var snpSwitch: UISwitch!
    @IBOutlet weak var beginUpdateButton: UIButton!
    @IBOutlet weak var optionsStack: UIStackView!
    @IBOutlet weak var bottomMargin: NSLayoutConstraint!
    @IBOutlet weak var versionField: UITextField!
    @IBOutlet weak var addressField: UITextField!
    
    // MARK: - Variables
    var btDevice: CBPeripheral!
    var curFirmware: ImportedFirmwareImage!
    var firmwareManager = OADFirmwareManager()
    var firmwareImages = [ImportedFirmwareImage]()
    var firmwareDownloadView: StatusLogViewController!
    var waitWriteSem: DispatchSemaphore!
    var blockWriteSem: DispatchSemaphore!
    var waitForCommandSem: DispatchSemaphore!
    var metaDataRejected: Bool = false
    var writeCommandFailed: Bool = false
    var codeExampleTable: CodeExampleTableViewController!
    var curBlockSize: Int = FIRMWARE_OAD_MAX_PAYLOAD_SIZE
    var curPacketSize: Int = FIRMWARE_OAD_MAX_PAYLOAD_SIZE
    var connectionCon: DiscoveryViewController!
    var timeoutTimer: Timer!
    var referencedController: BLEDeviceTableViewController!
    var cancelInvoked: Bool = false
    let hexSet = NSCharacterSet(charactersIn: "0123456789ABCDEFabcdef").inverted
    
    // MARK: - UUIDs
    let TI_OAD_SERVICE_UUID = CBUUID(string: "F000FFC0-0451-4000-B000-000000000000")
    let TI_OAD_IMG_IDENTIFY_UUID = CBUUID(string: "F000FFC1-0451-4000-B000-000000000000")
    let TI_OAD_IMG_BLOCK_UUID = CBUUID(string: "F000FFC2-0451-4000-B000-000000000000")
    let TI_OAD_EXT_CTL_UUID = CBUUID(string: "F000FFC5-0451-4000-B000-000000000000")
    
    // MARK: - Bluetooth Characteristics
    var imgIdentChar: CBCharacteristic!
    var imgBlockChar: CBCharacteristic!
    var imgExtCtlChar: CBCharacteristic!
    
    // MARK: - NSCoding
    func loadFirmwareImages() -> [ImportedFirmwareImage]?
    {
        var firmwareImages = NSKeyedUnarchiver.unarchiveObject(withFile: ImportedFirmwareImage.ArchiveURL.path)
            as? [ImportedFirmwareImage]
        
        if(firmwareImages == nil)
        {
            firmwareImages = [ImportedFirmwareImage]()
        }
        
        return firmwareImages
    }
    
    func saveFirmwareImages()
    {
        NSKeyedArchiver.archiveRootObject(firmwareImages, toFile: ImportedFirmwareImage.ArchiveURL.path)
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        btDevice.delegate = self
        btDevice.discoverServices(nil)

        sourceLabel.layer.cornerRadius = 6.0
        firmwareOptionsViewLabel.layer.cornerRadius = 6.0
        
        firmwareSourceView.layer.cornerRadius = 6.0
        firmwareSourceView.layer.shadowOpacity = 0.7
        firmwareSourceView.layer.shadowRadius = 6.0
        firmwareSourceView.layer.shadowOffset = CGSize(width: 0.0,
                                                       height: 8.0)
        firmwareSourceView.layer.shadowColor = UIColor(displayP3Red: ShadowColor,
                                                       green: ShadowColor,
                                                       blue: ShadowColor,
                                                       alpha: 0.5).cgColor
        
        firmwareOptionsView.layer.cornerRadius = 6.0
        firmwareOptionsView.layer.shadowOpacity = 0.7
        firmwareOptionsView.layer.shadowRadius = 6.0
        firmwareOptionsView.layer.shadowOffset = CGSize(width: 0.0,
                                                       height: 8.0)
        firmwareOptionsView.layer.shadowColor = UIColor(displayP3Red: ShadowColor,
                                                       green: ShadowColor,
                                                       blue: ShadowColor,
                                                       alpha: 0.5).cgColor
        
        beginUpdateButton.clipsToBounds = true
        beginUpdateButton.layer.cornerRadius = 8.0
        
        firmwareImages = loadFirmwareImages()!
        
        /* Adding observer to shift view for keyboard */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(self.keyboardNotification),
                                               name: NSNotification.Name.UIKeyboardWillChangeFrame,
                                               object: nil)
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(OADViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
        addressField.delegate = self
        versionField.delegate = self
        
        sapLabel.adjustsFontSizeToFitWidth = true
        snpLabel.adjustsFontSizeToFitWidth = true
        
        viewRotated()
        
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    // MARK: - Actions
    @IBAction func snpButtonChanged(_ sender: UISwitch)
    {
        if(!sender.isOn)
        {
            sapLabel.font = UIFont.boldSystemFont(ofSize: sapLabel.font.pointSize)
            snpLabel.font = UIFont.systemFont(ofSize: snpLabel.font.pointSize)
        }
        else
        {
            snpLabel.font = UIFont.boldSystemFont(ofSize: snpLabel.font.pointSize)
            sapLabel.font = UIFont.systemFont(ofSize: sapLabel.font.pointSize)
        }
    }
    
    @IBAction func updateFirmwarePressed(_ sender: Any)
    {
        if(curFirmware == nil)
        {
            let alert = UIAlertController(title: "No Firmware Selected",
                                          message: "Please select firmware before initiated update.",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        if((addressField.text?.isEmpty)! || (versionField.text?.isEmpty)!)
        {
            let alert = UIAlertController(title: "Invalid Values",
                                          message: "Invalid values for version/address",
                                          preferredStyle: UIAlertControllerStyle.alert)
            alert.addAction(UIAlertAction(title: "OK", style: UIAlertActionStyle.default, handler: nil))
            self.present(alert, animated: true, completion: nil)
            return
        }
        
        versionField.resignFirstResponder()
        addressField.resignFirstResponder()
        UIApplication.shared.isIdleTimerDisabled = true
        
        self.cancelInvoked = false
        self.performSegue(withIdentifier: "startDownloadSegue", sender: self)
        
    }
    
    // MARK: - Functions
    
    /* Called when the firmware update window is dismissed */
    func completionForUpdate()
    {
        if(self.writeCommandFailed == true || self.metaDataRejected == true)
        {
            
        }
        else
        {
            _ = self.navigationController?.popToViewController(self.codeExampleTable, animated: true)
        }
        
        UIApplication.shared.isIdleTimerDisabled = false
    }
    
    /* Initiates the firmware update with the current settings */
    func initiateFirmwareUpdate()
    {
        
        var firmwareTarget: FirmwareTarget!
        var firmwareFormat: FirmwareFormat!
        let firmwareArray: [UInt8]
        let metadataArray: [UInt8]
        
        if(snpSwitch.isOn)
        {
            firmwareTarget = .FIRMWARE_CC2650SNP
        }
        else
        {
            firmwareTarget = .FIRMWARE_MSP432SAP
        }
        
        if(curFirmware.filename?.contains(".txt"))!
        {
            firmwareFormat = .FIRMWARE_FORMAT_TITXT
        }
        else
        {
            firmwareFormat = .FIRMWARE_FORMAT_HEX
        }
        
        /* Initializing the semaphore we will use for synchronization between the BLE communication */
        waitWriteSem = DispatchSemaphore(value: 0)
        blockWriteSem = DispatchSemaphore(value: 0)
        waitForCommandSem = DispatchSemaphore(value: 0)
        
        /* Resetting variables */
        metaDataRejected = false
        writeCommandFailed = false
        
        /* Parsing the firmware out of the current image */
        (firmwareArray, metadataArray) = firmwareManager.prepareTXTFile(rawData: curFirmware.firmwareImageArray!,
                                                                        firmwareType: firmwareFormat,
                                                                        firmwareTarget: firmwareTarget,
                                                                        version: versionField.text!,
                                                                        address: addressField.text!)

        DispatchQueue.global(qos: .userInitiated).async
        {
            self.sendFirmwareToDevice(metaImage: metadataArray, firmwareImage: firmwareArray)

            DispatchQueue.main.async
            {
                if(self.cancelInvoked)
                {
                    
                }
                else if(self.writeCommandFailed == true || self.metaDataRejected == true)
                {
                    self.firmwareDownloadView.printLogMessage(message: "Firmware update completed with errors!")
                }
                else
                {
                    self.firmwareDownloadView.printLogMessage(message: "Firmware update completed successfully!")
                }
                
                self.firmwareDownloadView.doneButton.titleLabel?.text = "Done"
            }
        }
    }
    
    func cancelUpdate()
    {
        self.cancelInvoked = true
        waitWriteSem.signal()
        blockWriteSem.signal()
        waitForCommandSem.signal()
    }
    
    func sendFirmwareToDevice(metaImage: [UInt8], firmwareImage: [UInt8])
    {
        /* Setting the block size on the device */
        let blockSizeCommand = [UInt8]([FIRMWARE_OAD_BLK_SIZE_CMD,
                                        UInt8(FIRMWARE_OAD_MAX_PAYLOAD_SIZE >> 8),
                                        UInt8(FIRMWARE_OAD_MAX_PAYLOAD_SIZE & 0xFF)])
        self.btDevice.writeValue(Data(blockSizeCommand), for: imgExtCtlChar,
                                 type: CBCharacteristicWriteType.withoutResponse)
        waitForCommandSem.wait()
        
        if(self.cancelInvoked)
        {
            return
        }
     
        /* Send the metadata to the device */
        let imageSize: UInt32 = fromByteArray([UInt8](metaImage[6..<10]), UInt32.self)
        let totalBlocks = Int(ceil((Double(imageSize*4))/Double(FIRMWARE_OAD_MAX_PAYLOAD_SIZE)))
        
        firmwareDownloadView.printLogMessage(message: "Firmware update initialized.")
        firmwareDownloadView.printLogMessage(message: "Meta Data Info:")
        firmwareDownloadView.printLogMessage(message: "    CRC32: " + String(format: "0x%X",
            fromByteArray([UInt8](metaImage[0..<4]), UInt32.self)))
        firmwareDownloadView.printLogMessage(message: "    Version: " + String(format: "0x%X",
            fromByteArray([UInt8](metaImage[4..<6]), UInt16.self)))
        firmwareDownloadView.printLogMessage(message: "    Length: \(imageSize)")
        firmwareDownloadView.printLogMessage(message: "    Address: " + String(format: "0x%X",
            fromByteArray([UInt8](metaImage[10..<14]), UInt32.self)))
        firmwareDownloadView.printLogMessage(message: "    Image type: " + String(format: "0x%X", metaImage[14]))
        firmwareDownloadView.printLogMessage(message: "    Status: " + String(format: "0x%X", metaImage[15]))
        
        self.btDevice.writeValue(Data(metaImage), for: imgIdentChar, type: CBCharacteristicWriteType.withResponse)
        
        /* Waiting for write to clear */
        self.waitWriteSem.wait()
        
        if(self.cancelInvoked)
        {
            return
        }
        
        firmwareDownloadView.printLogMessage(message: "Performing firmware update...")
        
        /* Writing all of the data to the device */
        for ii in 0..<totalBlocks
        {
            if(self.cancelInvoked)
            {
                return
            }
            
            /* Waiting for the block request */
            blockWriteSem.wait()
            
            if(self.cancelInvoked)
            {
                return
            }
            
            /* If this variable is set it means the meta-data was rejected */
            if(metaDataRejected == true)
            {
                firmwareDownloadView.printLogMessage(message: "ERROR: Meta-data rejected! Aborting update.")
                return
            }
            else if(self.writeCommandFailed == true)
            {
                firmwareDownloadView.printLogMessage(message: "ERROR: Write command failed!")
                return
            }
            
            /* Calculating which chunk of the array to send */
            let curStartIndex = Int(ii * self.curBlockSize)
            var curEndIndex = Int()
            
            if((curStartIndex + self.curBlockSize) < firmwareImage.count)
            {
                curEndIndex = curStartIndex + self.curBlockSize
            }
            else
            {
                curEndIndex = firmwareImage.count
            }
            
            /* Constructing the packet */
            var curSendPacket = [UInt8](toByteArray(UInt32(ii)))
            curSendPacket.append(contentsOf: [UInt8](firmwareImage[curStartIndex..<curEndIndex]))
            
            while(curSendPacket.count < self.curPacketSize)
            {
                curSendPacket.append(0xFF)
            }
            
            /* Sending the packet */
            firmwareDownloadView.updateProgressBar(progress: Float(ii+1)/Float(totalBlocks))
            
            self.btDevice.writeValue(Data(curSendPacket), for: imgBlockChar,
                                     type: CBCharacteristicWriteType.withoutResponse)
            
        }
    }
    
    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            let thisService = service as CBService
            
            if(service.uuid == TI_OAD_SERVICE_UUID)
            {
                peripheral.discoverCharacteristics(nil, for: thisService)
                return
            }
        }
    }
    
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        if service.uuid == TI_OAD_SERVICE_UUID
        {
            for characteristic in service.characteristics!
            {
                if(characteristic.uuid == TI_OAD_IMG_IDENTIFY_UUID)
                {
                    imgIdentChar = characteristic
                    btDevice.setNotifyValue(true, for: characteristic)
                }
                else if(characteristic.uuid == TI_OAD_IMG_BLOCK_UUID)
                {
                    imgBlockChar = characteristic
                    btDevice.setNotifyValue(true, for: characteristic)
                }
                else if(characteristic.uuid == TI_OAD_EXT_CTL_UUID)
                {
                    imgExtCtlChar = characteristic
                    btDevice.setNotifyValue(true, for: characteristic)
                    referencedController.timeoutTimer.invalidate()
                    referencedController.connectionCon.dismiss(animated: true, completion: nil)
                    
                }
                else
                {
                    print("OAD: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
                    continue
                }
            }
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        var values = [UInt8](repeating:0, count:characteristic.value!.count)
        characteristic.value?.copyBytes(to: &values,
                                        count: characteristic.value!.count)
        
        if(characteristic.uuid == TI_OAD_IMG_IDENTIFY_UUID)
        {
            /* Signaling to send the next block */
            metaDataRejected = true
            blockWriteSem.signal()
        }
        else if(characteristic.uuid == TI_OAD_IMG_BLOCK_UUID)
        {
            /* Signaling to send the next block */
            blockWriteSem.signal()
        }
        else if(characteristic.uuid == TI_OAD_EXT_CTL_UUID)
        {
            /* We are expecting a command for setting the block size */
            if(values[0] == FIRMWARE_OAD_BLK_SIZE_CMD)
            {
                var responseBlockSize = Int((values[2])) << 8
                responseBlockSize  = (responseBlockSize) | Int(values[1])
                firmwareDownloadView.printLogMessage(message: "Block size set to \(responseBlockSize)bytes")
                self.curBlockSize = responseBlockSize
                self.curPacketSize = responseBlockSize + 4
                
                waitForCommandSem.signal()
                
            }
        }
        else
        {
            print("OAD: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
        }
    }
    
    /* Debug log for enabling/disabling notifications */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateNotificationStateFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if(error != nil)
        {
            print("OAD ERROR: Enabling/disabling notifications failed \n ----- \(error.debugDescription)")
        }
        
        if(characteristic.isNotifying)
        {
            print("OAD: Notifications enabled for: \(characteristic.uuid.uuidString)")
        }
        else
        {
            print("OAD: Notifications disabled for: \(characteristic.uuid.uuidString)")
        }
    }

    /* Callback for writing a value */
    func peripheral(_ peripheral: CBPeripheral,
                    didWriteValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if(error != nil)
        {
            writeCommandFailed = true
            waitWriteSem.signal()
            blockWriteSem.signal()
        }
        else
        {
            waitWriteSem.signal()
        }
        
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                    self.optionsStack.isHidden = false
                }
                else if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.optionsStack.isHidden = true
                }
        }
    }
    
    @objc func keyboardNotification(notification: NSNotification)
    {
        
        if let userInfo = notification.userInfo
        {
            let endFrame = (userInfo[UIKeyboardFrameEndUserInfoKey] as? NSValue)?.cgRectValue
            let duration:TimeInterval = (userInfo[UIKeyboardAnimationDurationUserInfoKey] as? NSNumber)?.doubleValue ?? 0
            let animationCurveRawNSN = userInfo[UIKeyboardAnimationCurveUserInfoKey] as? NSNumber
            let animationCurveRaw = animationCurveRawNSN?.uintValue ?? UIViewAnimationOptions.curveEaseInOut.rawValue
            let animationCurve:UIViewAnimationOptions = UIViewAnimationOptions(rawValue: animationCurveRaw)
            if (endFrame?.origin.y)! >= UIScreen.main.bounds.size.height
            {
                bottomMargin?.constant = 10.0
            }
            else
            {
                bottomMargin?.constant = endFrame?.size.height ?? 10.0
            }
            UIView.animate(withDuration: duration,
                           delay: TimeInterval(0),
                           options: animationCurve,
                           animations: { self.view.layoutIfNeeded() },
                           completion: nil)
        }
    }
    
    // MARK: - Navigation
    override func prepare(for segue: UIStoryboardSegue, sender: Any?)
    {
        if segue.identifier == "selectSourceSegue"
        {
            let oadSourceController:FirmwareSourceViewController =
                segue.destination as! FirmwareSourceViewController
            oadSourceController.referencedController = self
        }
        else if segue.identifier == "startDownloadSegue"
        {
            let firmwareDownloadController: StatusLogViewController =
                segue.destination as! StatusLogViewController
            firmwareDownloadController.referencedController = self
            self.firmwareDownloadView = firmwareDownloadController
        }
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        super.viewWillDisappear(animated)
        
        let viewControllers = self.navigationController?.viewControllers
        
        if(viewControllers?.index(of: self) == nil)
        {
            referencedController.btManager.scanForPeripherals(withServices: nil, options: nil)
            referencedController.btManager.cancelPeripheralConnection(self.btDevice)
        }
    }
    
    // MARK: - UITextFieldDelegate
    func textFieldShouldReturn(_ textField: UITextField) -> Bool
    {
        textField.resignFirstResponder()
        return true
    }
    
    func textFieldDidEndEditing(_ textField: UITextField)
    {
        
    }
    
    func textField(_ textField: UITextField, shouldChangeCharactersIn range: NSRange,
        replacementString string: String) -> Bool
    {
        
        let separatedChar = string.components(separatedBy: hexSet)
        let filteredNums = separatedChar.joined(separator: "")
        
        if (filteredNums != string)
        {
            return false
        }
        
        if(textField == versionField && (string.characters.count + (versionField.text?.characters.count)!) <= 4)
        {
            return true
        }
        else if(textField == addressField && (string.characters.count + (addressField.text?.characters.count)!) <= 8)
        {
            return true
        }
        
        return false
    }
}
