/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import CoreBluetooth
import AVFoundation

class ProjectZeroTabBarController: UITabBarController, CBPeripheralDelegate
{
    
    // MARK Device UUIDs
    let TI_PZ_LED_SERVICE_UUID = CBUUID(string: "F0001110-0451-4000-B000-000000000000")
    let TI_PZ_LED_CHAR_0_UUID = CBUUID(string: "F0001111-0451-4000-B000-000000000000")
    let TI_PZ_BUTTON_SERVICE_UUID = CBUUID(string: "F0001120-0451-4000-B000-000000000000")
    let TI_PZ_BUTTON_CHAR_0_UUID = CBUUID(string: "F0001121-0451-4000-B000-000000000000")
    let TI_PZ_BUTTON_CHAR_1_UUID = CBUUID(string: "F0001122-0451-4000-B000-000000000000")
    let TI_PZ_DATA_SERVICE_UUID = CBUUID(string: "F0001130-0451-4000-B000-000000000000")
    let TI_PZ_DATA_CHAR_0_UUID = CBUUID(string: "F0001131-0451-4000-B000-000000000000")
    let TI_PZ_DATA_CHAR_1_UUID = CBUUID(string: "F0001132-0451-4000-B000-000000000000")
    
    // MARK: Bluetooth Characteristics
    var ledCharLED0: CBCharacteristic!
    var buttonCharS1: CBCharacteristic!
    var buttonCharS2: CBCharacteristic!
    var dataCharString: CBCharacteristic!
    var dataCharData: CBCharacteristic!
   
    // MARK: Variables
    var btDevice: CBPeripheral!
    var referencedController: BLEDeviceTableViewController!
    var ledServiceController: LEDServiceViewController!
    var buttonServiceController: ButtonServiceViewController!
    var dataServiceController: DataServiceViewController!
    
    // MARK: CBPeripheralDelegate
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverServices error: Error?)
    {
        for service in peripheral.services!
        {
            let thisService = service as CBService
            peripheral.discoverCharacteristics(nil, for: thisService)
        }
    }
    
    func peripheral(_ peripheral: CBPeripheral,
                    didDiscoverCharacteristicsFor service: CBService,
                    error: Error?)
    {
        if service.uuid == TI_PZ_LED_SERVICE_UUID
        {
            for characteristic in service.characteristics!
            {
                if characteristic.uuid == TI_PZ_LED_CHAR_0_UUID
                {
                    ledCharLED0 = characteristic
                }
                else
                {
                    print("ProjectZero: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
                    continue
                }
                
                btDevice.readValue(for: characteristic)
            }
        }
        else if service.uuid == TI_PZ_BUTTON_SERVICE_UUID
        {
            for characteristic in service.characteristics!
            {
                if characteristic.uuid == TI_PZ_BUTTON_CHAR_0_UUID
                {
                    buttonCharS1 = characteristic
                }
                else if characteristic.uuid == TI_PZ_BUTTON_CHAR_1_UUID
                {
                    buttonCharS2 = characteristic
                }
                else
                {
                    print("ProjectZero: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
                    continue
                }
                
                btDevice.setNotifyValue(true, for: characteristic)
            }
        }
        else if service.uuid == TI_PZ_DATA_SERVICE_UUID
        {
            for characteristic in service.characteristics!
            {
                if characteristic.uuid == TI_PZ_DATA_CHAR_0_UUID
                {
                    dataCharString = characteristic
                }
                else if characteristic.uuid == TI_PZ_DATA_CHAR_1_UUID
                {
                    dataCharData = characteristic
                    btDevice.setNotifyValue(true, for: characteristic)
                    
                    if(ledServiceController != nil && ledServiceController.hasLoaded)
                    {
                        referencedController.connectionCon.dismiss(animated: true, completion: nil)
                        referencedController.timeoutTimer.invalidate()
                    }
                }
                else
                {
                    print("ProjectZero: Unknown characteristic UUID: \(characteristic.uuid.uuidString)")
                    continue
                }
            }
        }
        else
        {
            print("ProjectZero: Unknown Service UUID \(service.uuid.uuidString)")
        }
    }
    
    /* Callback for a characteristic read/notification */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateValueFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        var values = [UInt8](repeating:0, count:characteristic.value!.count)
        characteristic.value?.copyBytes(to: &values,
                                        count: characteristic.value!.count)
        
        if characteristic.uuid == TI_PZ_BUTTON_CHAR_0_UUID
        {
            if buttonServiceController == nil
            {
                return
            }
            
            if values[0] == 1
            {
                buttonServiceController.buttonS1.image = UIImage(named:"buttonon_s1")
                
                do
                {
                    buttonServiceController.s1EffectPlayer =
                        try AVAudioPlayer(contentsOf: buttonServiceController.s1SoundURL)
                    buttonServiceController.s1EffectPlayer.prepareToPlay()
                    buttonServiceController.s1EffectPlayer.play()
                }
                catch
                {
                    print("eLock: ERROR - Could not play effect")
                }
            }
            else
            {
                buttonServiceController.buttonS1.image = UIImage(named:"buttonoff_s1")
                buttonServiceController.s1EffectPlayer.stop()
            }
        }
        else if characteristic.uuid == TI_PZ_BUTTON_CHAR_1_UUID
        {
            if(buttonServiceController == nil)
            {
                return
            }
            
            if values[0] == 1
            {
                buttonServiceController.buttonS2.image = UIImage(named:"buttonon_s2")
                do
                {
                    buttonServiceController.s2EffectPlayer =
                        try AVAudioPlayer(contentsOf: buttonServiceController.s2SoundURL)
                    buttonServiceController.s2EffectPlayer.prepareToPlay()
                    buttonServiceController.s2EffectPlayer.play()
                }
                catch
                {
                    print("eLock: ERROR - Could not play effect")
                }
            }
            else
            {
                buttonServiceController.buttonS2.image = UIImage(named:"buttonoff_s2")
                buttonServiceController.s2EffectPlayer.stop()
            }
        }
        else if(characteristic.uuid == TI_PZ_DATA_CHAR_0_UUID)
        {
            if(dataServiceController == nil)
            {
                return
            }
            
            let readString = String(bytes: values, encoding: String.Encoding.utf8)
            
            if(readString != nil)
            {
                dataServiceController.printLogMessage(message: "Value of " + readString! + " read back.")
            }
        }
        else if(characteristic.uuid == TI_PZ_DATA_CHAR_1_UUID)
        {
            if(dataServiceController == nil)
            {
                return
            }
            
            var combinedString = ""
            
            for char in values
            {
                combinedString =  String(format: "%02hhx", char) + combinedString
            }
            
            combinedString = "0x" + combinedString
            
            
            dataServiceController.printLogMessage(message: "Value of " + combinedString + " notified")
            
        }
        else
        {
            print("ProjectZero: Invalid characteristic read (\(characteristic.uuid.uuidString))")
        }
        
    }
    
    /* Debug log for enabling/disabling notifications */
    func peripheral(_ peripheral: CBPeripheral,
                    didUpdateNotificationStateFor characteristic: CBCharacteristic,
                    error: Error?)
    {
        if(characteristic.isNotifying)
        {
            print("ProjectZero: Notifications enabled for: \(characteristic.uuid.uuidString)")
        }
        else
        {
            print("ProjectZero: Notifications disabled for: \(characteristic.uuid.uuidString)")
        }
    }
    
    /* Callback for writing a value */
    func peripheral(_ peripheral: CBPeripheral,
                    didWriteValueFor descriptor: CBDescriptor, error: Error?)
    {
        print("ProjectZero: Wrote descriptor value")
    }
    
    func peripheral(_ peripheral: CBPeripheral, didWriteValueFor characteristic: CBCharacteristic, error: Error?)
    {
        if(error != nil)
        {
            print("ProjectZero: Wrote value with error: " +  error.debugDescription)

        }
        else
        {
            /* If it is the data service, let's read back the characteristic */
            if(characteristic.uuid == TI_PZ_DATA_CHAR_0_UUID)
            {
                btDevice.readValue(for: dataCharString)
            }
        }
    }

    // MARK: Navigation
    override func viewDidLoad()
    {
        super.viewDidLoad()
    }

    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        referencedController.btManager.scanForPeripherals(withServices: nil, options: nil)
        referencedController.btManager.cancelPeripheralConnection(self.btDevice)
    }


}
