/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import NetworkExtension

class WiFiDirectProvisionViewController: UIViewController
{
    // MARK: Properties
    @IBOutlet weak var completeButton: UIButton!
    @IBOutlet weak var logArea: UITextView!
    @IBOutlet weak var logBox: UILabel!
    @IBOutlet weak var instructionBox: UILabel!
    @IBOutlet weak var doneBarItem: UIBarButtonItem!
    @IBOutlet weak var cancelBarItem: UIBarButtonItem!
    @IBOutlet weak var progressView: UIProgressView!
    
    
    // MARK: Variables
    var curSettings = WiFiSettings()
    let dateFormatter = DateFormatter()
    let provisionManager = WiFiProvisioningManager()
    
    // MARK: Actions
    @IBAction func completeButtonPressed(_ sender: UIButton)
    {
        if(completeButton.titleLabel?.text == "Cancel")
        {
            if(provisionManager.provisioningInProgress)
            {
                provisionManager.cancelProvisioning()
            }

            completeButton.setTitle("Retry", for: .normal)
        }
        else if(completeButton.titleLabel?.text == "Retry")
        {
            cancelBarItem.isEnabled = false
            doneBarItem.isEnabled = false
            completeButton.setTitle("Cancel", for: .normal)
            startProvisioning()
        }
        else if(completeButton.titleLabel?.text == "Done")
        {
            let provDevice = provisionManager.curProvisionedDevice
            let currentNet = WiFiProvisioningManager.currentSSIDs()
            let routerName = provDevice.wifiSettings.routerName
            var routerDev: NEHotspotConfiguration?
            
            if(currentNet.count == 0 || currentNet[0] != routerName)
            {
                /* Creating a configuration for the SimpleLink device we want to connect to */
                if(provDevice.wifiSettings.routerSecType == WiFiSecurityType.None)
                {
                    routerDev = NEHotspotConfiguration(ssid: routerName)
                }
                else if(provDevice.wifiSettings.routerSecType == WiFiSecurityType.WPA)
                {
                    routerDev = NEHotspotConfiguration(ssid: routerName,
                                                           passphrase: provDevice.wifiSettings.routerKey,
                                                           isWEP: false)
                }
                else if(provDevice.wifiSettings.routerSecType == WiFiSecurityType.WEP)
                {
                    routerDev = NEHotspotConfiguration(ssid: routerName,
                                                           passphrase: provDevice.wifiSettings.routerKey,
                                                           isWEP: true)
                }
                
                NEHotspotConfigurationManager.shared.apply(routerDev!, completionHandler:
                { (error) in
                    self.completeProvisioning()
                })
            }
            else
            {
                completeProvisioning()
            }
        }
    }
    
    @IBAction func cancelBarItemPressed(_ sender: Any)
    {
        self.navigationController?.popViewController(animated: true)
    }
    
    @IBAction func doneBarPressed(_ sender: Any)
    {
        self.navigationController?.popViewController(animated: true)
    }
    
    // MARK: Functions
    func printLogMessage(message: String)
    {
        DispatchQueue.main.async
        {
            self.logArea.text = self.logArea.text + self.dateFormatter.string(from: Date()) + ":    " + message + "\n"
            let range = NSRange(location: self.logArea.text.characters.count-1,length: 0)
            self.logArea.scrollRangeToVisible(range)
        }
    }
    
    func completeProvisioning()
    {
        UIApplication.shared.open(URL(string: provisionManager.curProvisionedDevice.commonURL)!,
                                  options: [:], completionHandler: nil)
        self.navigationController?.popToRootViewController(animated: true)
    }
    
    override func viewDidLoad()
    {
        super.viewDidLoad()
        
        dateFormatter.dateFormat = "HH:mm:ss"
        completeButton.layer.cornerRadius = 8.0
        completeButton.clipsToBounds = true
        logBox.layer.cornerRadius = 8.0
        logBox.clipsToBounds = true
        logArea.clipsToBounds = true
        logArea.layer.cornerRadius = 8.0
        provisionManager.referenceController  = self
        provisionManager.wifiSettings = self.curSettings
        
        /* Adding observer to rotate view */
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(WiFiDirectProvisionViewController.viewRotated),
                                               name: NSNotification.Name.UIDeviceOrientationDidChange, object: nil)
        
        startProvisioning()
    }
    
    func startProvisioning()
    {
        progressView.setProgress(0.0, animated: false)
        DispatchQueue.global().async
            {
                if(self.provisionManager.provisionDevice() == true)
                {
                    DispatchQueue.main.async
                    {
                        self.doneBarItem.isEnabled = true
                        self.completeButton.titleLabel?.text = "Done"
                    }
                }
                else
                {
                    DispatchQueue.main.async
                    {
                        self.cancelBarItem.isEnabled = true
                        self.completeButton.setTitle("Retry", for: .normal)
                    }
                }
            }
    }
    
    override func viewWillAppear(_ animated: Bool)
    {
        UIApplication.shared.isIdleTimerDisabled = true
    }
    
    override func viewWillDisappear(_ animated: Bool)
    {
        /* Removing observer to center the text */
        super.viewWillDisappear(animated)
        UIApplication.shared.isIdleTimerDisabled = false
    }
    override func didReceiveMemoryWarning()
    {
        super.didReceiveMemoryWarning()
    }
    
    @objc func viewRotated()
    {
        DispatchQueue.main.async
            {
                if UIDeviceOrientationIsPortrait(UIDevice.current.orientation)
                {
                   self.instructionBox.isHidden = false
                   self.completeButton.isHidden = false
                }
                
                if UIDeviceOrientationIsLandscape(UIDevice.current.orientation)
                {
                    self.instructionBox.isHidden = true
                    self.completeButton.isHidden = true
                }
        }
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?,
                               context: UnsafeMutableRawPointer?)
    {
        let textView = object as! UITextView
        var topCorrect = (textView.bounds.size.height - textView.contentSize.height * textView.zoomScale) / 2
        topCorrect = topCorrect < 0.0 ? 0.0 : topCorrect;
        textView.contentInset.top = topCorrect
    }
}
