/*
 * Copyright (c) 2017, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import UIKit
import Alamofire
import CocoaAsyncSocket
import NetworkExtension
import SystemConfiguration.CaptiveNetwork

class WiFiProvisioningManager : NSObject, GCDAsyncUdpSocketDelegate
{
    // MARK: Variables
    var productVersion: String = ""
    var referenceController: WiFiDirectProvisionViewController!
    var wifiSettings: WiFiSettings!
    var resolvedURL = ""
    var urlBase = "http://"
    var ipAddress = ""
    var udpSocket: GCDAsyncUdpSocket!
    var semaphore: DispatchSemaphore!
    var provisioningInProgress = false
    var curProvisionedDevice = ProvisionedWiFiDevice()
    var timeoutTimer: Timer!
    
    /* Customized Alamofire session manager that lets us use untrusted certificates */
    private static var simplelinkAFManager: Alamofire.SessionManager =
    {
        /* Apply customer server trust policies */
        let serverTrustPolicies: [String: ServerTrustPolicy] =
        [
            "mysimplelink.net": .disableEvaluation
        ]
        
        let configuration = URLSessionConfiguration.default
        configuration.httpAdditionalHeaders = Alamofire.SessionManager.defaultHTTPHeaders
        let manager = Alamofire.SessionManager(configuration: URLSessionConfiguration.default,
                                               serverTrustPolicyManager: ServerTrustPolicyManager(policies: serverTrustPolicies))
        
        /* Note that the code below disables verification and allows us to use self-signed certificates. In an actual
             application you should emphatically not do this and only use a properly signed certificate. Again,
             IN A REAL APPLICATION DO NOT DO THIS! */
        manager.delegate.sessionDidReceiveChallenge = {session, challenge in
            var disposition: URLSession.AuthChallengeDisposition = .performDefaultHandling
            var credential: URLCredential?
            
            if challenge.protectionSpace.authenticationMethod == NSURLAuthenticationMethodServerTrust,
                let trust = challenge.protectionSpace.serverTrust
            {
                disposition = URLSession.AuthChallengeDisposition.useCredential
                credential = URLCredential(trust: trust)
            }
            else
            {
                if challenge.previousFailureCount > 0
                {
                    disposition = .cancelAuthenticationChallenge
                } else
                {
                    credential = manager.session.configuration.urlCredentialStorage?.defaultCredential(for: challenge.protectionSpace)
                    
                    if credential != nil
                    {
                        disposition = .useCredential
                    }
                }
            }
            return (disposition, credential)
        }
        
        return manager
    }()
    
    // MARK: Functions
    static func currentSSIDs() -> [String]
    {
        guard let interfaceNames = CNCopySupportedInterfaces() as? [String] else
        {
            return []
        }
        return interfaceNames.flatMap
        { name in
            guard let info = CNCopyCurrentNetworkInfo(name as CFString) as? [String:AnyObject] else
            {
                return nil
            }
            guard let ssid = info[kCNNetworkInfoKeySSID as String] as? String else
            {
                return nil
            }
            return ssid
        }
    }
    
    func resetVariables()
    {
        productVersion = ""
        resolvedURL = ""
        urlBase = "http://"
        ipAddress = ""
        provisioningInProgress = false
        curProvisionedDevice = ProvisionedWiFiDevice()
        DispatchQueue.main.async
        {
            self.updateProgressBar(progress: 0.0)
        }
    }
    
    
    func provisionDevice() -> Bool
    {
        let devInfo = WIFI_DEFAULT_URL + WIFI_PARAM_FILE
        var queryURL = ""
        let afManager = WiFiProvisioningManager.simplelinkAFManager
        var simplelinkDev: NEHotspotConfiguration?
        var routerDev: NEHotspotConfiguration?
        var resCode = true
        var numOfTries = 0
        var currentDevice = ""
        var timeout = DispatchTime.now()
        let dateFormatter = DateFormatter()
        
        /* Intializaing Variables */
        resetVariables()
        curProvisionedDevice.wifiSettings = wifiSettings
        semaphore = DispatchSemaphore(value: 0)
        dateFormatter.dateStyle = .long
        dateFormatter.timeStyle = .short
        
        /* Marking the provisioning in progress */
        provisioningInProgress = true
        updateProgressBar(progress: 0.0)
        
        /* Getting list of connected SSIDs so we can prepopulate the AP router field */
        let currentNet = WiFiProvisioningManager.currentSSIDs()
        if(currentNet.count > 0)
        {
            currentDevice = currentNet[0]
        }
        
        /* Creating a configuration for the SimpleLink device we want to connect to */
        if(wifiSettings.devSecType == WiFiSecurityType.None)
        {
            simplelinkDev = NEHotspotConfiguration(ssid: wifiSettings.simplelinkName)
        }
        else if(wifiSettings.devSecType == WiFiSecurityType.WPA)
        {
            simplelinkDev = NEHotspotConfiguration(ssid: wifiSettings.simplelinkName,
                                               passphrase: wifiSettings.simplelinkKey,
                                               isWEP: false)
        }
        else if(wifiSettings.devSecType == WiFiSecurityType.WEP)
        {
            simplelinkDev = NEHotspotConfiguration(ssid: wifiSettings.simplelinkName,
                                                   passphrase: wifiSettings.simplelinkKey,
                                                   isWEP: true)
        }
        
        /* If we are already connected to the device, we do not need to connect again. If we are not connected to
             the device we should try to connect to it. */
        if(currentDevice != wifiSettings.simplelinkName)
        {
            printLogMessage(message: "Connecting to network " + wifiSettings.simplelinkName + "...")
            
            /* Grabs the shared configuration manager and applies the connection of the SimpleLink device.*/
            NEHotspotConfigurationManager.shared.apply(simplelinkDev!, completionHandler:
            { (error) in
                
                /* In the completion handler we want to see if the current device is the one we actually tried to
                    connect to. If not we should return an error. */
                let connectedNet = WiFiProvisioningManager.currentSSIDs()
                if(error == nil && connectedNet.count > 0 && connectedNet[0] == self.wifiSettings.simplelinkName)
                {
                    currentDevice = connectedNet[0]
                }
                else
                {
                    currentDevice = ""
                }
                
                self.semaphore.signal()
            })

            /* Wait for our completion handler. The API that connects to the hotspot will automatically time out so
                 there is no need for a timed semaphore here. */
            semaphore.wait()
            
            /* If we did not actually connect to the network return an error. */
            if(provisioningInProgress == false || currentDevice == "")
            {
                printLogMessage(message: "Error connecting to network " + wifiSettings.simplelinkName + "!")
                return false
            }
            else
            {
                printLogMessage(message: "Connected to network " + wifiSettings.simplelinkName + "!")
            }
        }
        else
        {
            printLogMessage(message: "Already connected to network: " + currentDevice)
        }
        updateProgressBar(progress: 0.1)
        
        /* Getting the device version from the SimpleLink device. It takes around 30 seconds from the time we connect
            to the device to when we actually can read the device version... hence the repeated tries with the
             built in delay
         */
        printLogMessage(message: "Requesting device information from URL:\n" + devInfo)
        printLogMessage(message: "This could take up to sixty seconds for the device to be recognized in iOS.")
        
        DispatchQueue.main.async
        {
            self.timeoutTimer = Timer.scheduledTimer(timeInterval: 90.0, target: self, selector: #selector(self.timerTimedOut),
                                                userInfo: nil, repeats: false)
        }
        
        while(self.productVersion.isEmpty && numOfTries < WIFI_MAX_TRIES)
        {
            sleep(1)
            DispatchQueue.main.async
            {
                self.updateProgressBar(progress: self.referenceController.progressView.progress + 0.01)
            }
            /* Posting a HTTP GET to the SimpleLink device to get the version string. */
            afManager.request(devInfo, method: .get).responseString
                { response in
                    if(response.result.isSuccess)
                    {
                        /* Once we actually have the response we want to parse out the resolved URL and save the
                              relevant date. */
                        self.printLogMessage(message: "URL Response: " + (response.response?.url?.absoluteString)!)
                        self.printLogMessage(message: "Request Successful! Device version: " + response.result.value!)
                        self.productVersion = response.result.value!
                        self.resolvedURL = (response.response?.url?.absoluteString)!.replacingOccurrences(
                                                                                    of: WIFI_PARAM_FILE, with: "")
                        if(self.resolvedURL.starts(with: "https"))
                        {
                            self.urlBase = "https://"
                        }
                        else
                        {
                            self.urlBase = "http://"
                        }
                    }
                    else
                    {
                        numOfTries = numOfTries + 1
                    }
                    self.semaphore.signal()
            }
            
            /* Wait for completion of connection. Timeout is handled by the max number of tries parameter */
            semaphore.wait()
            
            if(provisioningInProgress == false || !timeoutTimer.isValid)
            {
                numOfTries = WIFI_MAX_TRIES
            }
        }
        
        if(self.timeoutTimer.isValid)
        {
            self.timeoutTimer.invalidate()
        }
        
        /* If we maxed out on tries it likely means we are not connected to a SimpleLink network */
        if(numOfTries == WIFI_MAX_TRIES)
        {
            self.printLogMessage(message: "Could not determine SimpleLink device version!")
            self.provisioningInProgress = false
            return false
        }
        updateProgressBar(progress: 0.4)
        curProvisionedDevice.directURL = self.resolvedURL
        
        /* Setting the device name of the SimpleLink device. */
        self.printLogMessage(message: "Changing device name to " + wifiSettings.devName)
        
        /* Accounting for the product version and setting the URL accordingly */
        if(self.productVersion == WIFI_R2_VERSION)
        {
            queryURL = self.resolvedURL + WIFI_R2_SETNAME_PATH
            curProvisionedDevice.devVersion = "Gen2"
        }
        else
        {
            // TODO: Account for gen1 devices here.
            curProvisionedDevice.devVersion = "Gen1"
        }
        
        /* Setting up the parameters for the device name. See SWRA513 for detailes */
        var params: Parameters = [WIFI_SETNAME_PARAM : wifiSettings.devName]
        
        /* Posting the request to set the device name */
        afManager.request(queryURL, method: .post, parameters: params).validate().responseString
        { response in
            if(response.result.isSuccess)
            {
                self.printLogMessage(message: "SimpleLink device name changed to " + self.wifiSettings.devName + ".")
            }
            else
            {
                self.printLogMessage(message: "SimpleLink device name could not be changed!")
                resCode = false
            }
            self.semaphore.signal()
        }
        /* Waiting for the response */
        semaphore.wait()
        
        /* If an error happened, removed the configured profile and return false */
        if(provisioningInProgress == false || resCode == false)
        {
            removeConfiguredProfiles()
            return false
        }
        updateProgressBar(progress: 0.5)
        
        /* Sending the router profile to the device. This will write the flash profile on the SimpleLink device. */
        self.printLogMessage(message: "Adding router profile to device")

        /* Setting up the parameters for adding the router profile. See SWRA513 for detailes */
        params =
        [
                WIFI_SSID_PARAM : wifiSettings.routerName,
                WIFI_SECOTHER_PARAM : "0",
        ]
        
        if(wifiSettings.routerSecType == WiFiSecurityType.WPA)
        {
            params[WIFI_SECTYPE_PARAM] =  WIFI_SEC_TYPE_WPA
            params[WIFI_SECPASS_PARAM] =  wifiSettings.routerKey
        }
        else if(wifiSettings.routerSecType == WiFiSecurityType.WEP)
        {
            params[WIFI_SECTYPE_PARAM] =  WIFI_SEC_TYPE_WEP
            params[WIFI_SECPASS_PARAM] =  wifiSettings.routerKey
        }
        else
        {
            params[WIFI_SECTYPE_PARAM] = WIFI_SEC_TYPE_NONE
        }
        
        /* Setting up and executing the HTTP POST request */
        queryURL = self.resolvedURL + WIFI_R2_SETPROF_PATH
        afManager.request(queryURL, method: .post, parameters: params).responseString
            { response in
                if(response.result.isSuccess)
                {
                    self.printLogMessage(message: "Router profile set on target device!")
                }
                else
                {
                    self.printLogMessage(message: "Router profile could not be set on device!")
                    resCode = false
                }
                self.semaphore.signal()
        }
        updateProgressBar(progress: 0.6)
        
        /* Waiting for a response and returning if we had an error */
        semaphore.wait()
        
        if(provisioningInProgress == false || resCode == false)
        {
            removeConfiguredProfiles()
            return false
        }
        
        /* Advancing the state machine. This causes the profile that we just sent to the device to actually persist
             and tells the SimpleLink device to start advertising it's IP address over UDP */
        self.printLogMessage(message: "Adding router profile to device")
        
        queryURL = self.resolvedURL + WIFI_R2_MOVESTATE_PATH
        
        afManager.request(queryURL, method: .post, parameters: params).responseString
            { response in
                if(response.result.isSuccess)
                {
                    self.printLogMessage(message: "State machine advanced!")
                }
                else
                {
                    self.printLogMessage(message: "State machine could not be advanced!")
                    resCode = false
                }
                self.semaphore.signal()
        }
        
        /* Wait for response and return an error if something bad happened */
        semaphore.wait()
        if(provisioningInProgress == false || resCode == false)
        {
            return false
        }
        updateProgressBar(progress: 0.65)
        
        /* Connecting to the router specified. The idea is that we will connect to the "common" router, find out
            the IP address of the SimpleLink device (through its UDP broadcast), and then send a confirmation
             to the device in order to get everything completed with provisioning. */
        if(wifiSettings.routerSecType == WiFiSecurityType.None)
        {
            routerDev = NEHotspotConfiguration(ssid: wifiSettings.routerName)
        }
        else if(wifiSettings.routerSecType == WiFiSecurityType.WPA)
        {
            routerDev = NEHotspotConfiguration(ssid: wifiSettings.routerName,
                                                   passphrase: wifiSettings.routerKey,
                                                   isWEP: false)
        }
        else if(wifiSettings.routerSecType == WiFiSecurityType.WEP)
        {
            routerDev = NEHotspotConfiguration(ssid: wifiSettings.routerName,
                                                   passphrase: wifiSettings.routerKey,
                                                   isWEP: true)
        }
        
        printLogMessage(message: "Connecting to network " + wifiSettings.routerName + "...")
        
        NEHotspotConfigurationManager.shared.apply(routerDev!, completionHandler:
        { (error) in
            /* In the completion handler we want to see if the current device is the one we actually tried to
             connect to. If not we should return an error. */
            let connectedNet = WiFiProvisioningManager.currentSSIDs()
            if(error != nil || connectedNet.count == 0 || connectedNet[0] != self.wifiSettings.routerName)
            {
                resCode = false
            }
            self.semaphore.signal()
        })
        
        /* Wait for the response and return an error if something happened */
        semaphore.wait()
        if(provisioningInProgress == false || resCode == false)
        {
            printLogMessage(message: "Error connecting to network " + wifiSettings.routerName + "!")
            removeConfiguredProfiles()
            return false
        }
        else
        {
            printLogMessage(message: "Connected to network " + wifiSettings.routerName + "!")
        }
        updateProgressBar(progress: 0.7)
        
        /* Starting the UDP Socket to listen for the result. Once we connect to the device it will start to
         broadcast its IP address over UDP. We listen for this and then send a confirmation command that
         actually completes the provisioning process. */
        udpSocket = GCDAsyncUdpSocket(delegate: self, delegateQueue: DispatchQueue.main)
        
        printLogMessage(message: "Opening UDP socket and listening for broadcast...")
        do
        {
            try udpSocket.bind(toPort: WIFI_UDP_CONFIRM_PORT)
            try udpSocket.beginReceiving()
        }
        catch
        {
            printLogMessage(message: "Could not bind UDP socket for listening!")
            removeConfiguredProfiles()
            return false
        }
        
        /* Make the timeout for the UDP broadcast 30 seconds. If we can't get the IP address of the device through
             the "common" access point we should just connect to the device and confirm it directly */
        timeout = DispatchTime.now() + DispatchTimeInterval.seconds(30)
        if(semaphore.wait(timeout: timeout) == .timedOut)
        {
            udpSocket.close()
            
            self.printLogMessage(message: "Could not confirm through shared network. Confirming directly with device.")
            self.printLogMessage(message: "Connecting to " + wifiSettings.simplelinkName + "...")
            
            /* If we failed to receive any response from the device we should connect directly and verify the
                 communication was successful */
            NEHotspotConfigurationManager.shared.apply(simplelinkDev!, completionHandler:
            { (error) in
                /* In the completion handler we want to see if the current device is the one we actually tried to
                 connect to. If not we should return an error. */
                let connectedNet = WiFiProvisioningManager.currentSSIDs()
                if(error != nil || connectedNet.count == 0 || connectedNet[0] != self.wifiSettings.simplelinkName)
                {
                    resCode = false
                }
                self.semaphore.signal()
            })
            
            semaphore.wait()
            
            /* If we couldn't connect we should bail out altogether */
            if(provisioningInProgress == false || resCode == false)
            {
                self.printLogMessage(message: "Could not connect to " + wifiSettings.simplelinkName + "!")
                removeConfiguredProfiles()
                return false
            }
            else
            {
                curProvisionedDevice.confirmMethod = "Direct Connect"
                curProvisionedDevice.commonURL = "N/A"
            }
            
            self.printLogMessage(message: "Connected to " + wifiSettings.devName + "!")
        }
        else if(provisioningInProgress == false)
        {
            removeConfiguredProfiles()
            return false
        }
        updateProgressBar(progress: 0.8)
        
        /* Posting an HTTP get of the confirmation status. This HTTP get will finish the state machine on the SimpleLink
             device and complete the overall provisioning proccess */
        DispatchQueue.main.async
        {
            self.timeoutTimer = Timer.scheduledTimer(timeInterval: 90.0, target: self,
                                                     selector: #selector(self.timerTimedOut),
                                                     userInfo: nil, repeats: false)
        }
        
        self.printLogMessage(message: "Requesting confirmation status... ")
        queryURL = self.resolvedURL + WIFI_CONFIRM_REQUEST
        numOfTries = 0
        var responseCode = "-1"
        while(responseCode == "-1" && numOfTries < WIFI_MAX_TRIES)
        {
            sleep(1)
            afManager.request(queryURL, method: .get).validate().responseString
            { response in
                if(response.result.isSuccess)
                {
                    self.printLogMessage(message: "Request Successful! Confirmation Code: " + response.result.value!)
                    responseCode = response.result.value!
                }
                else
                {
                    numOfTries = numOfTries + 1
                }
                self.semaphore.signal()
            }
            
            /* Wait for the response */
            semaphore.wait()
            
            if(provisioningInProgress == false || !timeoutTimer.isValid)
            {
                numOfTries = WIFI_MAX_TRIES
            }
        }
        
        if(self.timeoutTimer.isValid)
        {
            self.timeoutTimer.invalidate()
        }
        
        /* Checking to see what the confirmation code is and making the relative prints */
        if(responseCode == "-1")
        {
            self.printLogMessage(message: "ERROR! Could not retrieve confirmation code!")
            resCode = false
        }
        else if(responseCode == WIFI_CONF_NO_IP)
        {
            self.printLogMessage(message: "ERROR! Could not obtain an IP address!")
            resCode = false
        }
        else if(responseCode == WIFI_CONF_NO_SSID)
        {
            self.printLogMessage(message: "ERROR! SimpleLink device could not find network " + wifiSettings.routerName)
            resCode = false
        }
        else if(responseCode == WIFI_CONF_CONN_FAIL)
        {
            self.printLogMessage(message: "ERROR! Connection failed! Maybe the security key is wrong?")
            resCode = false
        }
        else if(responseCode == WIFI_CONF_NOT_STARTED)
        {
            self.printLogMessage(message: "ERROR! WiFi confirmation not started. ")
            resCode = false
        }
        else if(responseCode == WIFI_CONF_GOOD)
        {
            self.printLogMessage(message: "Provisioning successfully completed without issue!")
            updateProgressBar(progress: 1.0)
            curProvisionedDevice.dateAdded = dateFormatter.string(from: Date())
            curProvisionedDevice.provisionLog = referenceController.logArea.text
            curProvisionedDevice.devURL = self.resolvedURL
            var provisionedDevices = loadProvisionedDevices()!
            provisionedDevices.append(curProvisionedDevice)
            NSKeyedArchiver.archiveRootObject(provisionedDevices, toFile: ProvisionedWiFiDevice.ArchiveURL.path)
        }
        
        /* Returning the error code */
        removeConfiguredProfiles()
        provisioningInProgress = false
        return resCode
    }
    
    func removeConfiguredProfiles()
    {
        NEHotspotConfigurationManager.shared.removeConfiguration(forSSID: wifiSettings.simplelinkName)
    }
    
    func printLogMessage(message: String)
    {
        referenceController.printLogMessage(message: message)
    }
    
    func cancelProvisioning()
    {
        if(timeoutTimer != nil)
        {
            timeoutTimer.invalidate()
        }
        
        self.provisioningInProgress = false
        self.semaphore.signal()
    }
    
    func updateProgressBar(progress: Float)
    {
        DispatchQueue.main.async
            {
                self.referenceController.progressView.setProgress(progress, animated: true)
            }
    }
    
    @objc func timerTimedOut()
    {
        timeoutTimer.invalidate()
        semaphore.signal()
    }
    
    func getWiFiAddress() -> String?
    {
        var address : String?
        
        /* Get list of all interfaces on the local machine */
        var ifaddr : UnsafeMutablePointer<ifaddrs>?
        guard getifaddrs(&ifaddr) == 0 else { return nil }
        guard let firstAddr = ifaddr else { return nil }
        
        /* For each interface ... */
        for ifptr in sequence(first: firstAddr, next: { $0.pointee.ifa_next })
        {
            let interface = ifptr.pointee
            
            /* Check for IPv4 or IPv6 interface:c*/
            let addrFamily = interface.ifa_addr.pointee.sa_family
            if addrFamily == UInt8(AF_INET) || addrFamily == UInt8(AF_INET6)
            {
                
                /* Check interface name: */
                let name = String(cString: interface.ifa_name)
                if  name == "en0"
                {
                    
                    /* Convert interface address to a human readable string: */
                    var hostname = [CChar](repeating: 0, count: Int(NI_MAXHOST))
                    getnameinfo(interface.ifa_addr, socklen_t(interface.ifa_addr.pointee.sa_len),
                                &hostname, socklen_t(hostname.count),
                                nil, socklen_t(0), NI_NUMERICHOST)
                    address = String(cString: hostname)
                }
            }
        }
        freeifaddrs(ifaddr)
        
        return address
    }
    
    // MARK: - NSCoding
    func loadProvisionedDevices() -> [ProvisionedWiFiDevice]?
    {
        var provisionedDevices = NSKeyedUnarchiver.unarchiveObject(withFile: ProvisionedWiFiDevice.ArchiveURL.path)
            as? [ProvisionedWiFiDevice]
        
        if(provisionedDevices == nil)
        {
            provisionedDevices = [ProvisionedWiFiDevice]()
        }
        
        return provisionedDevices
    }
    
    // MARK: GCDAsyncUdpSocketDelegate
    func udpSocket(_ sock: GCDAsyncUdpSocket, didReceive data: Data, fromAddress address: Data,
                   withFilterContext filterContext: Any?)
    {
        /* If we are receiving a packet and the socket is already closed treat this as an invalid read */
        if(udpSocket.isClosed())
        {
            return
        }
        
        /* Parsing out the data packet and seeing if it matches the format for our SimpleLink device */
        let stringData = NSString(data: data, encoding: String.Encoding.utf8.rawValue)
        let components = stringData?.components(separatedBy: ",")
        
        if(components != nil && components?.count == 2)
        {
            self.resolvedURL = self.urlBase + components![0]
            printLogMessage(message: "UDP Broadcast Found! \nAddress: " + self.resolvedURL + "\nDevice: "
                + components![1])
            curProvisionedDevice.confirmMethod = "UDP Common Broadcast"
            curProvisionedDevice.commonURL = self.resolvedURL

            udpSocket.close()
            semaphore.signal()
        }
    }
}
